/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.prognose;

import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.concurrent.TimeUnit;

import com.bitctrl.Constants;

import de.bsvrz.ibv.simbaustelle.parameter.IBssParameterObserver;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.ibv.simbaustelle.verwaltung.BaustellenSchnittstelle;
import de.bsvrz.ibv.simbaustelle.verwaltung.BssSchritt;
import de.bsvrz.ibv.simbaustelle.verwaltung.BssStau;
import de.bsvrz.ibv.simbaustelle.verwaltung.SimulationsAuftrag;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * berechnet die für eine Baustelle zu erwartetenden Staus.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public abstract class BaustellenSimulator {

	/**
	 * Logger für Debugausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die Baustelle, für die die Berechnung erfolgen soll.
	 */
	private final BaustellenSchnittstelle baustelle;

	/**
	 * die Parameter für die Berechnung.<br>
	 * Das Objekt wird aus der übergeordneten Baustellenverwaltung übernommen.
	 */
	private final IBssParameterObserver parameter;

	/** der Name des Auftraggebers der Simulation. */
	private final SimulationsAuftrag auftrag;

	/** die Liste der ermittelten Staus. */
	private final List<BssStau> stauListe = new ArrayList<>();

	/** die Liste der Schritte bei der Stauberechnung. */
	private List<BssSchritt> schrittListe;

	/**
	 * erzeugt eine Instanz der Klasse zur Berechnung von Staus in einer Baustelle.
	 *
	 * @param auftrag
	 *            der auszuführende Simulationsauftrag
	 * @param parameter
	 *            die Parameter für die Berechnung aus der Baustellenverwaltung
	 */
	public BaustellenSimulator(final SimulationsAuftrag auftrag, final IBssParameterObserver parameter) {

		LOGGER.finer("Erzeuge BaustellenSimulator für Baustelle: " + auftrag.getBaustelle().getName());

		this.baustelle = auftrag.getBaustelle();
		this.auftrag = auftrag;
		this.parameter = parameter;
	}

	public SimulationsAuftrag getAuftrag() {
		return auftrag;
	}

	public BaustellenSchnittstelle getBaustelle() {
		return baustelle;
	}

	/**
	 * Gibt den aktuellen Zeitpunkt der Berechnung als text zur&uuml;ck.
	 *
	 * @param zeit
	 *            Zeit in ms
	 * @return Zeitpunkt als Text
	 */
	private String getBerechnungZeitText(final long zeit) {
		return "Zeitpunkt: " + DateFormat.getDateTimeInstance().format(new Date(zeit));
	}

	public IBssParameterObserver getParameter() {
		return parameter;
	}

	public List<BssStau> getStauListe() {
		return stauListe;
	}

	public List<BssSchritt> getSchrittListe() {
		if( schrittListe == null) {
			return Collections.emptyList();
		}
		return Collections.unmodifiableList(schrittListe);
	}

	/**
	 * ermittelt die Liste der Staus, die innerhalb des Prognosezeitraums zu
	 * erwarten sind.
	 *
	 * @param zeitRaum
	 *            der Zeitraum, für den die Staus prognostiziert werden sollen
	 * @param zufVerkehr
	 *            der zufließende Verkehr
	 * @throws UndefiniertException
	 *             ein für die Prognose erforderlicher Wert konnte nicht ermittelt
	 *             werden, die Prognose wurde nicht ausgeführt.
	 */
	protected final void prognoseErstellen(final BerechnungsZeitRaumSchnittstelle zeitRaum,
			final ZufliessenderVerkehrSchnittstelle zufVerkehr) throws UndefiniertException {

		final long stBeginn = parameter.getLaengeStauBeginn();
		final long stAufhebung = parameter.getLaengeStauAufhebung();
		long aktuelleLaenge = 0;
		long fzImStau = 0;
		BssStau aktuellerStau = null;

		stauListe.clear();
		schrittListe = new ArrayList<>(zeitRaum.getIntervalle().size() + 1);

		final double schrittweite = (double) zeitRaum.getSchrittweiteInMsec() / TimeUnit.HOURS.toMillis(1);

		LOGGER.finer(baustelle.getName() + ": Beginne Stauberechnung mit " + " Parameter 'LängeStaubeginn': " + stBeginn
				+ ", 'LängeStauAufhebung': " + stAufhebung + ", 'Schrittweite': " + schrittweite);

		for (final long zs : zeitRaum.getIntervalle()) {
			try {
				final double zufluss = zufVerkehr.getFzg(zs);
				Long restKapazitaet = auftrag.getRestKapazitaet();
				if (restKapazitaet == null) {
					restKapazitaet = baustelle.getDaten().getRestKapazitaet(zs);
				}

				LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs) + ": Zufluss: " + zufluss
						+ ", Restkapazität: " + restKapazitaet);

				final double fzDiff = (zufluss - restKapazitaet) * schrittweite;

				final int fahrstreifenAnzahl = baustelle.anzahlFahrStreifen(aktuelleLaenge);
				final long laengePkw = parameter.getLaengePkw();

				final long lDiff = (long) ((fzDiff / fahrstreifenAnzahl) * laengePkw);
				aktuelleLaenge = Math.max(0, aktuelleLaenge + lDiff);
				fzImStau = Math.max(0, fzImStau + (long) fzDiff);
				final long verlustZeit = (long) (((double) fzImStau / restKapazitaet)
						* Constants.MILLIS_PER_HOUR);

				schrittListe.add(new BssSchritt(zs - zeitRaum.getSchrittweiteInMsec(), aktuelleLaenge, verlustZeit, fzImStau));

				LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs) + ": FzDiff: " + fzDiff
						+ ", FzImStau: " + fzImStau + ", Verlustzeit: " + verlustZeit);

				if (aktuellerStau == null) {
					if (aktuelleLaenge > stBeginn) {
						LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs)
								+ ": neuen Stau erzeugen: Länge: " + aktuelleLaenge + ", Verlustzeit: " + verlustZeit);
						aktuellerStau = new BssStau(zs, aktuelleLaenge, verlustZeit);
						stauListe.add(aktuellerStau);
					} else {
						LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs)
								+ ": es wird kein Stau erzeugt, da ermittelte Staulänge (" + aktuelleLaenge
								+ "m) < Parameter 'LängeStaubeginn' (" + stBeginn + "m)");
					}
				} else {
					if (aktuelleLaenge < stAufhebung) {
						LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs) + ": Stau beenden: Länge ("
								+ aktuelleLaenge + "m) < Parameter 'LängeStauAufhebung' (" + stAufhebung + "m)");
						aktuellerStau.beenden(zs);
						aktuellerStau = null;
					} else {
						LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs)
								+ ": Stau aktualisieren: Länge: " + aktuelleLaenge + ", Verlustzeit: " + verlustZeit);
						aktuellerStau.aktualisieren(zs, aktuelleLaenge, verlustZeit);
					}
				}
			} catch (final UndefiniertException ex) {
				if (!BaustellenSimulation.getInstance().isAllowGaps()) {
					throw new UndefiniertException("Daten für den Prognosezeitraum sind unvollständig: " + ex.getLocalizedMessage(), ex);
				}
			}
		}
	}
}
